from Bio import SeqIO
import numpy as np
import pathlib

from primalscheme3.core.bedfiles import read_in_bedprimerpairs, BedPrimerPair


MATCH_LIMIT = 15


def ref_fkmer_gen(ref_seq: str, stop_index: int):
    """
    Yeilds base by base fkmer sequences from a reference sequence
    """
    for i in range(stop_index, -1, -1):
        # Skip over gaps
        if ref_seq[i] == "-":
            continue
        yield ref_seq[i]


def fkmer_map(fkmer_seq: set[str], ref_seq: str) -> dict[int, list[int]]:
    """Calculates the index which the fkmer maps to in the MSA

    Args:
        fkmer_seq (str): The fkmer sequence
        msa_array (np.ndarray): The raw referance sequence
        start_index (int): The fixed index of the fkmer in the MSA. Non inclusive

    Returns:
        int: The fkmer stop index with the closest match using hamming distance
    """
    # Get the length of the fkmer seq

    # Key: index, Value: number of mismatches for each fkmer seq
    match_dict: dict[int, list[int]] = {}

    fkmer_max_len = max([len(x) for x in fkmer_seq])

    # Find the best match
    for stopindex in range(fkmer_max_len, len(ref_seq)):
        match_dict[stopindex] = []
        for fseq in fkmer_seq:
            # Calculate the pseudo hamming distance
            matches = 0
            for fseqbase, refbase in zip(fseq[::-1], ref_fkmer_gen(ref_seq, stopindex)):
                if fseqbase == refbase:
                    matches += 1
            # Append the mismatches
            match_dict[stopindex].append(matches)

    # Find the best scoreing index
    return match_dict


def ref_rkmer_gen(ref_seq: str, start_index: int):
    """
    Yeilds base by base rkmer sequences from a reference sequence
    """
    for i in range(start_index, len(ref_seq)):
        # Skip over gaps
        if ref_seq[i] == "-":
            continue
        yield ref_seq[i]


def rkmer_map(rkmer_seq: set[str], ref_seq: str) -> dict[int, list[int]]:
    """Calculates the index which the rkmer maps to in the MSA

    Args:
        rkmer_seq (str): The rkmer sequence
        msa_array (np.ndarray): The raw referance sequence
        start_index (int): The fixed index of the rkmer in the MSA. Non inclusive

    Returns:
        int: The rkmer start index with the closest match using hamming distance
    """
    # Get the length of the rkmer seq

    # Key: index, Value: number of mismatches for each rkmer seq
    match_dict: dict[int, list[int]] = {}

    rkmer_max_len = max([len(x) for x in rkmer_seq])

    # Find the best match
    for startindex in range(0, len(ref_seq) - rkmer_max_len):
        match_dict[startindex] = []
        for rseq in rkmer_seq:
            # Calculate the pseudo hamming distance
            matches = 0
            for rseqbase, refbase in zip(rseq, ref_rkmer_gen(ref_seq, startindex)):
                if rseqbase == refbase:
                    matches += 1
            # Append the mismatches
            match_dict[startindex].append(matches)

    # Find the best scoreing index
    return match_dict


def remap_func(msa: pathlib.Path, bedfile: pathlib.Path, refid: str):
    # Read in the bedfile
    primerpairs: list[BedPrimerPair] = read_in_bedprimerpairs(bedfile)
    # If there is more than one MSA in the bedfile, raise an error
    if len(set([primerpair.chromname for primerpair in primerpairs])) > 1:
        raise ValueError(
            "Primers for more than one MSA is present in the bedfile, cannot remap"
        )

    # Find which reference genome should be used
    input_genomes = SeqIO.index(msa, "fasta")
    # Check reference genome is in the MSA
    if refid not in input_genomes:
        raise ValueError(
            f"The reference genome with ID '{refid}' is not present in the MSA"
        )
    else:
        # Get the reference sequence
        ref_seq = str(input_genomes[refid].seq)  # type: ignore
        ref_seq = ref_seq.replace("-", "")

    # Use a hamming distance score to map the primers.
    # As we are not sure of the indexing system
    for primerpair in primerpairs:
        # Keep track of which primers have been mapped
        fprimer_mapped = False
        rprimer_mapped = False

        # For each fkmer
        fkmer_matches = fkmer_map(primerpair.fprimer.seqs, ref_seq)

        # Look for best matches
        fkmer_best_matches = {
            index: max(scores)
            for index, scores in fkmer_matches.items()
            if max(scores) >= MATCH_LIMIT
        }
        if len(fkmer_best_matches) == 0:
            primerpair.fprimer.end = 0
            fprimer_mapped = True
        elif len(fkmer_best_matches) == 1:
            # Remap the fkmer
            fkmer_stop_index = list(fkmer_best_matches.keys())[0]
            primerpair.fprimer.end = fkmer_stop_index
            fprimer_mapped = True
        else:
            # If there are multiple best matches, pick the best
            fkmer_best_match: int = max(fkmer_best_matches.values())
            # Check best is unique
            fkmer_best_match_index = [
                index
                for index, score in fkmer_best_matches.items()
                if score == fkmer_best_match
            ]
            if len(fkmer_best_match_index) == 1:
                # Remap the fkmer
                primerpair.fprimer.end = fkmer_best_match_index[0]
                fprimer_mapped = True
            else:
                fprimer_mapped = False  # reassign to False for clarity

        # For each rkmer
        rkmer_matches = rkmer_map(primerpair.rprimer.seqs, ref_seq)

        # Look for best matches
        rkmer_best_matches = {
            index: max(scores)
            for index, scores in rkmer_matches.items()
            if max(scores) >= MATCH_LIMIT
        }
        if len(rkmer_best_matches) == 0:
            primerpair.rprimer.start = 0
            rprimer_mapped = True
        elif len(rkmer_best_matches) == 1:
            # Remap the rkmer
            rkmer_start_index = list(rkmer_best_matches.keys())[0]
            primerpair.rprimer.start = rkmer_start_index
            rprimer_mapped = True
        else:
            # If there are multiple best matches, pick the best
            rkmer_best_match: int = max(rkmer_best_matches.values())
            # Check best is unique
            rkmer_best_match_index = [
                index
                for index, score in rkmer_best_matches.items()
                if score == rkmer_best_match
            ]
            if len(rkmer_best_match_index) == 1:
                # Remap the rkmer
                primerpair.rprimer.start = rkmer_best_match_index[0]
                rprimer_mapped = True
            else:
                rprimer_mapped = False

        # Use both primer positions to resolve ambiguity
        if rprimer_mapped == False and fprimer_mapped == False:
            print(
                f"WARNING: The primer pair {primerpair.amplicon_number} could not be mapped"
            )
        elif fprimer_mapped and rprimer_mapped == False:
            # Use the fprimer mapping to resolve the ambiguity
            closest_mapping = min(
                rkmer_best_match_index, key=lambda x: (x - primerpair.fprimer.end) * -1
            )
            primerpair.rprimer.start = closest_mapping
        elif rprimer_mapped and fprimer_mapped == False:
            # Use the rprimer mapping to resolve the ambiguity
            closest_mapping = min(
                fkmer_best_match_index,
                key=lambda x: (primerpair.rprimer.start - x) * -1,
            )
            primerpair.fprimer.end = closest_mapping

        # Check the primer pair is valid
        if primerpair.fprimer.end > primerpair.rprimer.start:
            print(f"The primer pair {primerpair.amplicon_number} is invalid")

        print(primerpair.to_bed(primerpair.chromname, primerpair.ampliconprefix))



import pathlib


def read_bed_file(bedfilepath: pathlib.Path) -> tuple[list[list[str]], list[str]]:
    """
    Reads a bed file and returns a list of lists of strings.

    :return: bedfile_list, bedfile_header
    """

    with open(bedfilepath, "r") as bedfile:
        bedfile_list: list[list[str]] = []
        bedfile_header: list[str] = []

        for line in bedfile:
            # Header line
            if line.startswith("#"):
                bedfile_header.append(line.strip())
            if line:  # If not empty
                bedfile_list.append(line.strip().split("\t"))

    return bedfile_list, bedfile_header



from Bio import SeqIO
import numpy as np
import pathlib

from primalscheme3.core.bedfiles import read_in_bedprimerpairs, BedPrimerPair


def consensus_remap(msa: pathlib.Path, bedfile: pathlib.Path, refid: str):
    # Read in the bedfile
    primerpairs: list[BedPrimerPair] = read_in_bedprimerpairs(bedfile)
    # If there is more than one MSA in the bedfile, raise an error
    if len(set([primerpair.chromname for primerpair in primerpairs])) > 1:
        raise ValueError(
            "Primers for more than one MSA is present in the bedfile, cannot remap"
        )

    # Find which reference genome should be used
    input_genomes = SeqIO.index(msa, "fasta")
    # Check reference genome is in the MSA
    if refid not in input_genomes:
        raise ValueError(
            f"The reference genome with ID '{refid}' is not present in the MSA"
        )
    else:
        # Get the reference sequence
        ref_seq = str(input_genomes[refid].seq)  # type: ignore

    # Create a lookup array for the MSA indexing to the referance genome
    mapping_array: list[None | int] = [None] * len(ref_seq)

    ref_index = 0
    for msa_index in range(len(ref_seq)):
        # If gap, skip
        if ref_seq[msa_index] == "-":
            continue
        else:
            mapping_array[msa_index] = ref_index
            ref_index += 1

    # Mapp each primerpair
    for primerpair in primerpairs:
        new_end = mapping_array[primerpair.fprimer.end]
        mapped = False

        if new_end is None:
            # Walk left until a non-gap is found
            new_index = primerpair.fprimer.end
            while new_index >= 0:
                new_end = mapping_array[new_index]
                # If valid index, break
                if new_end is not None:
                    primerpair.fprimer.end = new_end
                    mapped = True
                    break
                else:
                    new_index -= 1

            # Ensure all primers are in the index
            if not mapped:
                primerpair.fprimer.end = 0 + max(
                    [len(x) for x in primerpair.fprimer.seqs]
                )

        else:
            primerpair.fprimer.end = new_end

        new_start = mapping_array[primerpair.rprimer.start]
        if new_start is not None:
            primerpair.rprimer.start = new_start
        else:
            # Walk right until a non-gap is found
            new_index = primerpair.rprimer.start
            mapped = False
            while new_index < len(mapping_array):
                new_start = mapping_array[new_index]
                # If valid index, break
                if new_start is not None:
                    primerpair.rprimer.start = new_start
                    mapped = True
                    break
                else:
                    new_index += 1

            if not mapped:
                primerpair.rprimer.start = len(ref_seq.replace("-", "")) - max(
                    [len(x) for x in primerpair.rprimer.seqs]
                )

        primerpair.chromname = refid
        print(primerpair, end="")



import numpy as np
from Bio import SeqIO, Seq
import pathlib


def remove_terminal_n_func(msa: pathlib.Path, outpath: pathlib.Path):
    """
    Replaces the terminal Ns in the MSA with gaps
    """
    records_index = SeqIO.parse(msa, "fasta")

    with open(outpath, "w") as outfile:
        for record in records_index:
            seq_list: list[str] = list(str(record.seq).upper())
            seq_len = len(seq_list)

            # Solves the 5' end
            for col_index in range(0, seq_len):
                if seq_list[col_index] == "N" or seq_list[col_index] == "-":
                    seq_list[col_index] = "-"
                else:
                    break

            # Sovles the 3' end
            for rev_col_index in range(seq_len - 1, 0, -1):
                if seq_list[rev_col_index] == "N" or seq_list[rev_col_index] == "-":
                    seq_list[rev_col_index] = "-"
                else:
                    break

            # Recreate a Seq object
            seq = Seq.Seq("".join(seq_list))
            record.seq = seq
            # Write out the record
            SeqIO.write(record, outfile, "fasta")



import pathlib

from primalhelper.bedfile import read_bed_file


def get_amplicon_number(primerpair: list[str]) -> int:
    return int(primerpair[3].split("_")[1])


def primer_name_prefix(primerpair: list[str]) -> str:
    return "_".join(primerpair[3].split("_")[:-1])


def legacy_convert(bedfilepath: pathlib.Path, include_circular: bool = False):
    # Read in the bedfile
    primerpairs, _headers = read_bed_file(bedfilepath)

    # Get all reference names
    reference_names = set(pp[0] for pp in primerpairs)
    # Get all amplicon numbers
    amplicon_numbers = set(get_amplicon_number(pp) for pp in primerpairs)

    for reference_name in reference_names:
        primerpairs_for_reference = [
            pp for pp in primerpairs if pp[0] == reference_name
        ]

        # For each amplicon number, get all primerpairs
        for amplicon_number in amplicon_numbers:
            primerpairs_for_amplicon = [
                pp
                for pp in primerpairs_for_reference
                if get_amplicon_number(pp) == amplicon_number
            ]

            # Get the forward and reverse primers
            forward_primers = [fp for fp in primerpairs_for_amplicon if fp[5] == "+"]
            reverse_primers = [rp for rp in primerpairs_for_amplicon if rp[5] == "-"]

            # Get the furthest indexes
            fp_start_indexes = min([int(fp[1]) for fp in forward_primers])
            fp_end_indexes = max([int(fp[2]) for fp in forward_primers])
            rp_start_indexes = min([int(rp[1]) for rp in reverse_primers])
            rp_end_indexes = max([int(rp[2]) for rp in reverse_primers])

            # FPrimer Stem
            fp_name = primer_name_prefix(forward_primers[0])
            rp_name = primer_name_prefix(reverse_primers[0])

            # fp pool
            fp_pool = forward_primers[0][4]
            rp_pool = reverse_primers[0][4]

            # longest_seq
            fp_longest_seq = max([fp[6] for fp in forward_primers], key=len)
            rp_longest_seq = max([rp[6] for rp in reverse_primers], key=len)

            if include_circular:
                # Print Fp
                print(
                    f"{reference_name}\t{fp_start_indexes}\t{fp_end_indexes}\t{fp_name}\t{fp_pool}\t+\t{fp_longest_seq}"
                )
                # Print rp
                print(
                    f"{reference_name}\t{rp_start_indexes}\t{rp_end_indexes}\t{rp_name}\t{rp_pool}\t+\t{rp_longest_seq}"
                )
            elif fp_start_indexes < rp_start_indexes:  # Filter out circular primers
                # Print Fp
                print(
                    f"{reference_name}\t{fp_start_indexes}\t{fp_end_indexes}\t{fp_name}\t{fp_pool}\t-\t{fp_longest_seq}"
                )
                # Print rp
                print(
                    f"{reference_name}\t{rp_start_indexes}\t{rp_end_indexes}\t{rp_name}\t{rp_pool}\t-\t{rp_longest_seq}"
                )


if __name__ == "__main__":
    legacy_convert(
        bedfilepath=pathlib.Path(
            "/Users/kentcg/primerschemes-fork/primerschemes/hbv/500/v1.1.0/primer.bed"
        ),
        include_circular=True,
    )



import pathlib
import typer
from typing_extensions import Annotated
from enum import Enum
from Bio import SeqIO, Seq, SeqRecord

# Import module
from primalhelper.__init__ import __version__
from primalhelper.hamming_remap import remap_func
from primalhelper.consensus_remap import consensus_remap
from primalhelper.legacy import legacy_convert
from primalhelper.remove_terminal_n import remove_terminal_n_func


# Create the typer app
app = typer.Typer()


class RemapMethod(Enum):
    hamming = "hamming"
    consensus = "consensus"


@app.command()
def remap(
    msa: Annotated[
        pathlib.Path,
        typer.Option(help="The path to the scheme directory", readable=True),
    ],
    bedfile: Annotated[
        pathlib.Path, typer.Option(help="The path to the bedfile", readable=True)
    ],
    refid: Annotated[str, typer.Option(help="The ID of the reference genome")],
    RemapMethod: Annotated[
        RemapMethod, typer.Option(help="The ID of the reference genome")
    ] = RemapMethod.consensus,
):
    """
    Remaps the given multiple sequence alignment (MSA) using the specified method.

    Args:
        msa (pathlib.Path): The path to the scheme directory.
        bedfile (pathlib.Path): The path to the bedfile.
        refid (str): The ID of the reference genome.
        RemapMethod (RemapMethod, optional): The remap method to use. Defaults to RemapMethod.consensus.

    Raises:
        NotImplementedError: If the specified remap method is not yet stable.

    """
    match RemapMethod:
        case RemapMethod.consensus:
            consensus_remap(msa, bedfile, refid)
        case RemapMethod.hamming:
            raise NotImplementedError("Hamming remap is not yet stable")
            remap_func(msa, bedfile, refid)


@app.command()
def legacy(
    bedfile: Annotated[
        pathlib.Path, typer.Option(help="The path to the bedfile", readable=True)
    ],
    include_circular: Annotated[
        bool, typer.Option(help="Include circular primers")
    ] = False,
):
    """
    Convert a new V2 style bedfile to V1 format for use with fieldbioinfomatics.

    Args:
        bedfile (pathlib.Path): The path to the bedfile.
    """
    legacy_convert(bedfile, include_circular)


@app.command()
def merge_segments(
    reference: Annotated[
        list[pathlib.Path],
        typer.Option(help="The referance genomes to merge", readable=True),
    ],
    bedfile: Annotated[pathlib.Path, typer.Option(help="The path to the bedfile")],
    output: Annotated[
        pathlib.Path,
        typer.Option(help="The dir to the output file"),
    ],
):
    spacer_len = 400
    fasta_list: list[str] = []

    for ref in reference:
        fasta_list.append(SeqIO.read(ref, "fasta").seq)

    # Merge the fasta
    merged_fasta = Seq.Seq(
        ("N" * spacer_len).join(fasta_list),
    )
    # Write out the fasta
    SeqIO.write(SeqRecord.SeqRecord(merged_fasta, id="merged_ref"), output, "fasta")


@app.command()
def remove_terminal_n(
    msa: Annotated[
        pathlib.Path,
        typer.Option(help="The path to the scheme directory", readable=True),
    ],
    output: Annotated[
        pathlib.Path, typer.Option(help="The path to the output file", writable=True)
    ],
):
    remove_terminal_n_func(msa, output)


if __name__ == "__main__":
    app()
